from grapevne.helpers import (
    init,
    script,
    resource,
    input,
    output,
    log,
    env,
    param,
    params,
)
from unittest import mock
from pathlib import Path
import pytest


class Workflow:
    def __init__(self, config):
        self.config = config


def test_script():
    init()
    with mock.patch(
        "grapevne.helpers.helpers.Helper._workflow_path",
        lambda self, path: Path("workflows") / path,
    ):
        assert script("script.py") == Path("workflows/scripts/script.py")


def test_resource():
    init()
    with mock.patch(
        "grapevne.helpers.helpers.Helper._workflow_path",
        lambda self, path: Path("workflows") / path,
    ):
        assert resource("resource.txt") == Path("workflows/../resources/resource.txt")


def test_input_single():
    workflow = Workflow(
        {
            "input_namespace": "in",
        }
    )
    init(workflow)
    assert Path(input("infile.txt")) == Path("results/in/infile.txt")


def test_input_multi():
    workflow = Workflow(
        {
            "input_namespace": {
                "port1": "in1",
                "port2": "in2",
            },
        }
    )
    init(workflow)
    assert Path(input("infile1.txt", "port1")) == Path("results/in1/infile1.txt")
    assert Path(input("infile2.txt", "port2")) == Path("results/in2/infile2.txt")


def test_output():
    workflow = Workflow(
        {
            "output_namespace": "out",
        }
    )
    init(workflow)
    assert Path(output("outfile.txt")) == Path("results/out/outfile.txt")


def test_log():
    init()
    assert log("rule.log") == "logs/rule.log"


def test_env():
    init()
    assert env("conda.yaml") == "envs/conda.yaml"


def test_param():
    workflow = Workflow(
        {
            "params": {
                "param1": "value1",
                "param2": {
                    "param3": "value3",
                },
            },
        }
    )
    init(workflow)
    assert param("param1") == "value1"
    assert param("param2", "param3") == "value3"


def test_param_notfound():
    workflow = Workflow(
        {
            "params": {
                "param1": "value1",
            },
        }
    )
    init(workflow)
    with pytest.raises(ValueError):
        param("param2")


def test_params():
    workflow = Workflow(
        {
            "params": {
                "param1": "value1",
                "param2": {
                    "param3": "value3",
                },
            },
        }
    )
    init(workflow)
    assert params("param1") == "value1"
    assert params("param2", "param3") == "value3"


def test_params_notfound():
    workflow = Workflow(
        {
            "params": {
                "param1": "value1",
            },
        }
    )
    init(workflow)
    with pytest.raises(ValueError):
        params("param2")



from pathlib import Path
from dataclasses import dataclass
from packaging.version import Version
from contextlib import contextmanager

try:
    import snakemake

except ImportError:
    raise ImportError(
        "This library must be called from within a snakemake environment."
    )


# Retain workflow _helper


@dataclass
class Helper:
    snakemake_version = Version(snakemake.__version__)

    def __init__(self, workflow=None):
        self.workflow = workflow
        self.config = workflow.config if workflow else None

    def _check_config(self):
        if self.config is None:
            raise ValueError("Configuration not loaded")

    def _check_workflow(self):
        if self.workflow is None:
            raise ValueError("Workflow not loaded")

    # Utility functions to return the workflow file path

    def _workflow_path(self, path):
        if Helper.snakemake_version < Version("8"):
            return self._workflow_path_snakemake_7(path)
        else:
            return self._workflow_path_snakemake_8(path)

    def _workflow_path_snakemake_7(self, path):
        return snakemake.workflow.srcdir(path)

    def _workflow_path_snakemake_8(self, path):
        return Path(self.workflow.current_basedir) / path

    # Utility functions to return the remote file path

    def _get_remote_file_path(self, path):
        if Helper.snakemake_version < Version("8"):
            return self._get_remote_file_path_snakemake_7(path)
        else:
            return self._get_remote_file_path_snakemake_8(path)

    def _get_remote_file_path_snakemake_7(self, path):
        return snakemake.remote.AUTO.remote(path)

    def _get_remote_file_path_snakemake_8(self, path):
        return self.workflow._storage_registry(path)

    def _get_file_path(self, path):
        return self._workflow_path(path)

    # Utility functions to return a module file path

    def _module_path(self, base, path):
        """Return the path to a module file"""
        folder = f"{base}"
        module_name = self.config.get("output_namespace", None) if self.config else None
        folder += f"/{module_name}" if module_name else ""
        return f"{folder}/{path}"

    # File-type specialisations

    def script(self, relpath):
        """Return the path to a script file"""
        return self._get_file_path(Path("scripts") / relpath)

    def resource(self, relpath):
        """Return the path to a resource file"""
        return self._get_file_path(Path("../resources") / relpath)

    def remote(self, path):
        """Return the path to a remote file"""
        return self._get_remote_file_path(path)

    def input(self, path, port=None):
        """Return the path to an input file

        Args:
            path (str): The path to the input file
            port (str): The name of the input port (optional)
        """
        self._check_config()
        input_namespace = self.config.get("input_namespace", None)
        if isinstance(input_namespace, str):
            if port:
                raise ValueError("Attempting to index a non-indexable input namespace")
            indir = self.config["input_namespace"]
        elif isinstance(input_namespace, dict):
            indir = self.config["input_namespace"].get(port, None)
        else:
            raise ValueError(
                "Snakemake config error - Input namespace type not recognised"
            )
        if not indir:
            raise ValueError("Attempting to read from a non-configured input port")
        if path:
            return f"results/{indir}/{path}"
        else:
            return f"results/{indir}"

    def output(self, path=None):
        """Return the path to an output file"""
        self._check_config()
        outdir = self.config.get("output_namespace", None)
        if not outdir:
            raise ValueError("Output namespace not defined in config")
        if path:
            return f"results/{outdir}/{path}"
        else:
            return f"results/{outdir}"

    def log(self, path=None):
        """Return the path to a log file"""
        path = path if path else "log.txt"
        return self._module_path("logs", path)

    def env(self, path):
        """Return the path to an environment file"""
        return f"envs/{path}"

    def benchmark(self, path=None):
        """Return the path to a benchmark file"""
        path = path if path else "benchmark.tsv"
        return self._module_path("benchmarks", path)

    # Parameter indexing

    def params(self, *args):
        """Return the value of a parameter in the configuration

        Args:
            *args:  The path to the parameter in the configuration
                    Example: params("foo", "bar") will return the value of
                        config["params"]["foo"]["bar"]
                    If no arguments are provided, return the entire params dictionary
        """
        self._check_config()
        if len(args) == 0:
            return self.config.get("params", {})
        value = self.config.get("params", {})
        for arg in args:
            if not isinstance(value, dict):
                raise ValueError(f"Attempting to index a non-indexable value: {value}")
            if arg not in value:
                raise ValueError(f"Parameter not found: {arg}")
            value = value.get(arg, {})
        return value

    # Alias function - 'params' is the Snakefile directive, but 'param' is more
    # intuitive and consistent with other helper functions.
    def param(self, *args):
        return self.params(*args)


@contextmanager
def grapevne_helper(globals_dict):
    workflow = globals_dict.get("workflow", None)
    gv = Helper(workflow)

    globals_dict["script"] = gv.script
    globals_dict["resource"] = gv.resource
    globals_dict["input"] = gv.input
    globals_dict["output"] = gv.output
    globals_dict["log"] = gv.log
    globals_dict["env"] = gv.env
    globals_dict["param"] = gv.param
    globals_dict["params"] = gv.params

    try:
        yield gv
    finally:
        del globals_dict["script"]
        del globals_dict["resource"]
        del globals_dict["input"]
        del globals_dict["output"]
        del globals_dict["log"]
        del globals_dict["env"]
        del globals_dict["params"]


_helper = Helper()


def init(workflow=None):
    _helper.workflow = workflow
    _helper.config = workflow.config if workflow else None
    _helper.snakemake_version = Version(snakemake.__version__)


def script(relpath):
    """Return the path to a script file"""
    return _helper.script(relpath)


def resource(relpath):
    """Return the path to a resource file"""
    return _helper.resource(relpath)


def remote(path):
    """Return the path to a remote file"""
    return _helper.remote(path)


def input(path, port=None):
    """Return the path to an input file

    Args:
        path (str): The path to the input file
        port (str): The name of the input port (optional)
    """
    return _helper.input(path, port)


def output(path=None):
    """Return the path to an output file"""
    return _helper.output(path)


def log(path=None):
    """Return the path to a log file"""
    return _helper.log(path)


def env(path):
    """Return the path to an environment file"""
    return _helper.env(path)


def benchmark(path=None):
    """Return the path to a benchmark file"""
    return _helper.benchmark(path)


def param(*args):
    return _helper.param(*args)


def params(*args):
    return _helper.params(*args)



import logging
import sys

try:
    import grapevne
except ImportError:
    import ensurepip
    import subprocess

    ensurepip.bootstrap()
    subprocess.check_call(
        [sys.executable, "-m", "pip", "install", "--upgrade", "grapevne"]
    )
    try:
        import grapevne
    except ImportError:
        logging.error("Failed to install grapevne. Exiting.")
        sys.exit(1)


def import_grapevne(workflow=None, version=None):
    _grapevne = grapevne.install(version)
    if workflow:
        _grapevne.helpers.init(workflow)
    return _grapevne.helpers



from .helpers import Helper  # noqa: F401
from .helpers import grapevne_helper  # noqa: F401
from .helpers import init  # noqa: F401
from .helpers import script  # noqa: F401
from .helpers import resource  # noqa: F401
from .helpers import remote  # noqa: F401
from .helpers import input  # noqa: F401
from .helpers import output  # noqa: F401
from .helpers import log  # noqa: F401
from .helpers import env  # noqa: F401
from .helpers import benchmark  # noqa: F401
from .helpers import param  # noqa: F401
from .helpers import params  # noqa: F401



import sys
import pytest
import grapevne


def test_install_none():
    grapevne_helpers = grapevne.install(version=None)
    root_module = sys.modules[grapevne_helpers.__name__.split(".")[0]]
    assert root_module.__version__ == grapevne.__version__


def test_install_current():
    grapevne_helpers = grapevne.install(version="current")
    root_module = sys.modules[grapevne_helpers.__name__.split(".")[0]]
    assert root_module.__version__ == grapevne.__version__


def test_install_this_version():
    version = grapevne.__version__
    grapevne_helpers = grapevne.install(version=version)
    root_module = sys.modules[grapevne_helpers.__name__.split(".")[0]]
    assert root_module.__version__ == version


def test_install_nonsense():
    version = "nonsense"
    with pytest.raises(ValueError):
        grapevne.install(version=version)



import sys
import ensurepip
import subprocess
import importlib
import logging
from packaging.version import Version
from .version import __version__


def _install_package(package_name: str, version: Version, path: str = ""):
    version_str = version.public
    path = f"{package_name}_v{version_str}"

    ensurepip.bootstrap(upgrade=True)
    args = [
        sys.executable,
        "-m",
        "pip",
        "install",
        f'"{package_name}~={version_str}"',
    ]
    if path:
        args += [
            "-t",
            path,
        ]
    try:
        logging.info(f"Installing {package_name} v{version_str}...")
        logging.debug(f"Running: {' '.join(args)}")
        subprocess.run(*args, check=True)
    except subprocess.CalledProcessError as e:
        logging.warning(f"Global install failed: {e}. Retrying with --user.")
        try:
            subprocess.run([*args, "--user", package_name], check=True)
        except subprocess.CalledProcessError as e:
            logging.error(f"User install failed: {e}. Exiting.")
            sys.exit(1)


def _get_versioned_module(
    version: Version, module_name: str = "grapevne", path: str = ""
):
    _install_package(module_name, version, path)
    try:
        spec = importlib.util.spec_from_file_location(module_name, path)
        if spec is None:
            raise ImportError(f"Failed to load {module_name} from {path}.")
        module = importlib.util.module_from_spec(spec)
    except ImportError as e:
        logging.error(f"Failed to import {module_name}: {e}. Exiting.")
        sys.exit(1)
    return module


def install(version):
    if not version or (isinstance(version, str) and version == "current"):
        # Return the current module
        logging.info("No version requirements specified.")
        return sys.modules[__name__.split(".")[0]]
    required_version = Version(version)
    logging.info(f"Installing grapevne v{required_version}...")
    current_version = Version(__version__)
    logging.info(f"Current version: v{current_version}")
    if current_version < required_version:
        # Install the required version and return the module
        logging.info(f"Installing grapevne v{required_version}...")
        module = _get_versioned_module(required_version)
        return module
    else:
        # Return the current module
        logging.info(f"grapevne v{current_version} is up-to-date.")
        return sys.modules[__name__.split(".")[0]]



import nox


@nox.session(python=["3.9", "3.10", "3.11", "3.12"])
def snakemake7(session):
    """Test Snakemake 7"""
    session.install("snakemake<8", "pulp<2.8", "pytest", ".")
    session.run("pytest")
    session.run("bash", "tests/run_tests.sh", external=True)


@nox.session(python=["3.12"])
def snakemake8(session):
    """Test Snakemake 8"""
    session.install("snakemake<9", "pytest", ".")
    session.run("pytest")
    session.run("bash", "tests/run_tests.sh", external=True)
